import typing
import dataclasses
from uuid import uuid4
from django.http import JsonResponse
from rest_framework import status
from morest.core import MorestJSONEncoder


class Response(JsonResponse):
    def __init__(self, 
                    data: dict = None, 
                    status_code: int = 200, 
                    status = 'ok',
                    message: str = 'OK', 
                    error_details: dict = None, 
                **kwargs):
        from morest.middlewares.requestid import RequestID
        
        data = {
            "data": data,
            "status": status,
            "status_code": status_code,
            "request_id": RequestID.get(uuid4().hex),
            "message": message,
        }
        if not 200 <= status_code <= 299:
            data['error_details'] = error_details

        super().__init__(data, status=status_code, encoder=MorestJSONEncoder, **kwargs)

    @classmethod
    def validation_error(cls, error_details: dict, status_code: int = status.HTTP_400_BAD_REQUEST) -> "Response":
        return cls(
            status_code=status_code,
            status='validation_error',
            message='Request body or query params validation error',
            error_details=error_details
        )
    
    @classmethod
    def from_status(cls, status: typing.Literal['ok', 'fail', 'error']) -> "Response":
        status_to_code = {
            "ok": 200,
            "fail": 400,
            "error": 500
        }
        return cls(
            status_code=status_to_code[status],
            status=status,
            message=status.capitalize()
        )
    


from morest.api.base import Response


class BaseError(BaseException):
    def __init__(self, message: str = None, error_details: dict = None):
        self.message = message or self.message
        self.error_details = error_details or self.error_details
        super().__init__()

    status_code = 500
    message = 'error'
    status = 'error'
    error_details = None

    def to_response(self) -> Response:
        return Response(
            data=None,
            status_code=self.status_code,
            status=self.status,
            message=self.message,
            error_details=self.error_details,
        )




import contextvars
from uuid import uuid4
from django.http import HttpRequest, HttpResponse

RequestID = contextvars.ContextVar('request_id')

class RequestIDMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response

  def __call__(self, request: HttpRequest):
    request_id = request.headers.get('Request-ID', uuid4().hex)
    RequestID.set(request_id)
    response: HttpResponse = self.get_response(request)
    response.headers['Request-ID'] = request_id
    return response



import json
import uuid
import decimal
import datetime
import dataclasses


from django.utils.duration import duration_iso_string
from django.utils.functional import Promise
from django.utils.timezone import is_aware


class MorestJSONEncoder(json.JSONEncoder):
    """
    JSONEncoder subclass that knows how to encode date/time, decimal types, dataclasses, and
    UUIDs.
    """

    def default(self, o):
        # See "Date Time String Format" in the ECMA-262 specification.
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith("+00:00"):
                r = r.removesuffix("+00:00") + "Z"
            return r
        elif isinstance(o, datetime.date):
            return o.isoformat()
        elif isinstance(o, datetime.time):
            if is_aware(o):
                raise ValueError("JSON can't represent timezone-aware times.")
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r
        elif isinstance(o, datetime.timedelta):
            return duration_iso_string(o)
        elif isinstance(o, (decimal.Decimal, uuid.UUID, Promise)):
            return str(o)
        elif dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        else:
            return super().default(o)


import logging
from traceback import format_exc
from django.http import HttpRequest, HttpResponse
from django.conf import settings
from rest_framework.exceptions import APIException
from rest_framework.views import exception_handler
from morest.errors import BaseError, InternalError
from morest.api import Response


class ExceptionMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def _handle_exception(self, exc):
        if isinstance(exc, BaseError):
            return exc.to_response()
        elif isinstance(exc, APIException):
            return Response(
                status=exc.detail.code,
                status_code=exc.status_code,
                message=str(exc.detail),
                error_details=exc.get_full_details(),
            )
        else:
            exception = format_exc()
            logging.error(exception)
            if settings.DEBUG:
                return InternalError(error_details={
                    "exception": exception,
                }).to_response()

            return InternalError().to_response()
        
    def __call__(self, request: HttpRequest):
        try:
            response: HttpResponse = self.get_response(request)
            return response
        except BaseException as e:
            return self._handle_exception(e)


def DRFExceptionMiddleware(exc, context):
    response = exception_handler(exc, context)
    if exc is not None:
        return ExceptionMiddleware(None)._handle_exception(exc)
    return response


import typing
from morest.errors import NotFoundError
from django.core.exceptions import ValidationError
from django.http import Http404
from django.shortcuts import get_object_or_404 as _get_object_or_404

T = typing.TypeVar("T")

def get_objects_or_404(qs: T, with_error_details: bool = False, **filters) -> T:
    try:
        return _get_object_or_404(qs, **filters)
    except (TypeError, ValueError, ValidationError, Http404):    
        if with_error_details:
            raise NotFoundError.with_object_details(qs.__name__, filters)
        else:
            raise NotFoundError



from .base import BaseError
from rest_framework import status


class InternalError(BaseError):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    message = 'Internal server error'
    status = 'error'


class InsufficientBalanceError(BaseError):
    status_code = status.HTTP_402_PAYMENT_REQUIRED
    message = 'Insufficient balance, try to top up before commiting the action'
    status = 'insufficient_balance'

    @classmethod
    def with_balance_details(cls, current_balance: float, required_balance: float, **kwargs) -> "InsufficientBalanceError":
        return cls(
            error_details={
                "current_balance": current_balance,
                "required_balance": required_balance,
                **kwargs
            }
        )


class AlreadyExistsError(BaseError):
    status_code = status.HTTP_409_CONFLICT
    message = 'Data already exists'
    status = 'data_already_exists'

    @classmethod
    def with_object_details(cls, object: str, key: str, value: str) -> "AlreadyExistsError":
        return cls(
            message='{object} with {key} "{value}" already exists'.format(
                object=object, 
                key=key,
                value=value,
            ),
            error_details={
                "object": object,
                "key": key,
                "value": value,
            }
        )


class NotFoundError(BaseError):
    status_code = status.HTTP_404_NOT_FOUND
    message = 'Not found error'
    status = 'not_found'

    @classmethod
    def with_object_details(cls, object: str, filters: dict) -> "NotFoundError":
        return cls(
            message=f'{object} with {filters=} not found',
            error_details={
                "object": object,
                "filters": filters,
            }
        )
    




from rest_framework import serializers


def build_response_serializer(serializer) -> serializers.Serializer:
    class Response(serializers.Serializer):
        data = serializer()
        status = serializers.CharField()
        status_code = serializers.IntegerField()
        request_id = serializers.CharField()
        message = serializers.CharField()

        class Meta:
            ref_name = f"Response{serializer.__name__}"
    return Response


import math
import typing
from dataclasses import dataclass
from rest_framework import serializers
from django.db.models.manager import BaseManager
from django.db.models.query import QuerySet
from morest.errors import InternalError
from morest.core import get_queryset
T = typing.TypeVar("T")

@dataclass
class PaginatedSerializedData:
    rows: typing.List[typing.Union[dict, T]]
    rows_count: int
    pages_count: int


class PaginationSerializer(serializers.Serializer):
    rows_name = 'rows'

    page = serializers.IntegerField(required=False, min_value=1, write_only=True)
    limit = serializers.IntegerField(required=False, min_value=1, write_only=True)

    def paginate(self, qs: typing.Union[BaseManager[T], typing.List[T]], serializer: serializers.Serializer = None, **kwargs) -> PaginatedSerializedData:
        page = self.validated_data.get('page', 1)
        limit = self.validated_data.get('limit', 20)

        qs = get_queryset(qs)

        if isinstance(qs, list):
            total_rows_count = len(qs)
            rows = qs[(page-1)*limit:page*limit]
        elif isinstance(qs, QuerySet) or isinstance(qs, BaseManager):
            total_rows_count = qs.count()
            rows = qs.all()[(page-1)*limit:page*limit]
        else:
            raise InternalError
        
        obj_serializer = lambda obj: serializer().to_representation(obj, **kwargs) if serializer is not None else obj 

        data = PaginatedSerializedData(
            rows=[obj_serializer(x) for x in rows],
            rows_count=len(rows),
            pages_count=math.ceil(total_rows_count/limit)
        )
        data.__dataclass_fields__['rows'].name = self.rows_name
        setattr(data, self.rows_name, data.rows)
        return data




from django.http import HttpRequest
from django.db.models.query import QuerySet
from rest_framework.views import APIView
from rest_framework.serializers import Serializer
from morest.utils import PaginationSerializer
from morest.api import Response
from morest.core import get_queryset, swagger

class ListFilterView(APIView):
    queryset: QuerySet
    serializer: Serializer
    filter_serializer: Serializer

    def get_queryset(self, request: HttpRequest) -> QuerySet:
        return get_queryset(self.queryset)
    
    def get_filter_serializer(self, request: HttpRequest) -> Serializer:
        return self.filter_serializer
    
    def get_filters(self, request: HttpRequest, filters: dict):
        if issubclass(self.get_filter_serializer(request), PaginationSerializer):
            filters.pop("page", None)
            filters.pop("limit", None)
        return filters
    
    def filter_queryset(self, request: HttpRequest, qs: QuerySet, filters: dict) -> QuerySet:
        return qs.filter(**filters).all()
    
    def get_serializer(self, request: HttpRequest) -> Serializer:
        return self.serializer
    
    def paginate(self, request: HttpRequest, filter_serializer: PaginationSerializer, qs: QuerySet):
        serializer = self.get_serializer(request)
        return filter_serializer.paginate(qs=qs, serializer=serializer)
    
    def serialize(self, request: HttpRequest, qs: QuerySet):
        serializer = self.get_serializer(request)
        return [serializer().to_representation(x) for x in qs]
    
    def get_response_data(self, request: HttpRequest, qs: QuerySet, data: dict):
        if isinstance(data, PaginationSerializer):
            return self.paginate(request=request, filter_serializer=data, qs=qs)
        else:
            return self.serialize(request=request, qs=qs)
        
    def get(self, request: HttpRequest):
        qs = self.get_queryset(request)
        filter_serializer = self.get_filter_serializer(request)
        data = filter_serializer(data=request.GET)
        if not data.is_valid():
            return Response.validation_error(data.errors)
        
        filters = self.get_filters(request, data.validated_data.copy())
        qs = self.filter_queryset(request, qs, filters)
        return Response(self.get_response_data(request, qs, data))
    
    @classmethod
    def as_view(cls, **initkwargs):
        try:
            from drf_yasg.utils import swagger_auto_schema
            class viewcls(cls):
                @swagger_auto_schema(
                    query_serializer=cls.filter_serializer,
                    responses={
                        200: swagger.build_response_serializer(cls.get_serializer(cls, None))
                    }
                )
                def get(self, request):
                    return super().get(request)
            cls = viewcls
        except ImportError:
            pass
        return super().as_view(**initkwargs)
    
